import { __extends } from 'tslib';
import { FormControl, FormGroup, FormBuilder } from '@angular/forms';
import { classToClass, plainToClass } from 'class-transformer';
import { MetadataStorage, ValidationTypes, Validator, getFromContainer, validateSync } from 'class-validator';
import { cloneDeep, mergeWith } from 'lodash-es';
import 'reflect-metadata';
import { BehaviorSubject } from 'rxjs';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template TModel
 */
var  /**
 * @template TModel
 */
DynamicFormGroup = /** @class */ (function (_super) {
    __extends(DynamicFormGroup, _super);
    function DynamicFormGroup(factoryModel, fields) {
        var _this = _super.call(this, {}) || this;
        _this.factoryModel = factoryModel;
        _this.fields = fields;
        _this.customValidateErrors = new BehaviorSubject({});
        _this.fields = _this.onlyFields(_this.fields);
        return _this;
    }
    Object.defineProperty(DynamicFormGroup.prototype, "externalErrors", {
        get: /**
         * @return {?}
         */
        function () {
            return this._externalErrors;
        },
        set: /**
         * @param {?} externalErrors
         * @return {?}
         */
        function (externalErrors) {
            this._externalErrors = externalErrors;
            this.validate();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DynamicFormGroup.prototype, "validatorOptions", {
        get: /**
         * @return {?}
         */
        function () {
            return this._validatorOptions;
        },
        set: /**
         * @param {?} validatorOptions
         * @return {?}
         */
        function (validatorOptions) {
            this._validatorOptions = validatorOptions;
            this.validate();
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?} fields
     * @param {?=} validatorOptions
     * @return {?}
     */
    DynamicFormGroup.getClassValidators = /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?} fields
     * @param {?=} validatorOptions
     * @return {?}
     */
    function (factoryModel, fields, validatorOptions) {
        /** @type {?} */
        var validationMetadatas = getFromContainer(MetadataStorage).getTargetValidationMetadatas(factoryModel, '', validatorOptions && validatorOptions.groups ? validatorOptions.groups : undefined);
        /** @type {?} */
        var formGroupFields = {};
        /** @type {?} */
        var validator = new Validator();
        Object.keys(fields).filter(function (key) { return key.indexOf('__') !== 0; }).forEach(function (key) {
            /** @type {?} */
            var formGroupField = formGroupFields[key];
            if (formGroupField === undefined) {
                formGroupField = Array.isArray(fields[key]) ? fields[key] : [];
            }
            if (!Array.isArray(fields[key])) {
                if (fields[key] instanceof DynamicFormGroup) {
                    fields[key].object = fields[key].fields;
                    formGroupField.push(fields[key]);
                }
                else {
                    formGroupField.push(fields[key]);
                }
            }
            /** @type {?} */
            var conditionalValidations = [];
            validationMetadatas.forEach(function (validationMetadata) {
                if (validationMetadata.propertyName === key && validationMetadata.type === 'conditionalValidation') {
                    conditionalValidations.push(validationMetadata);
                }
            });
            validationMetadatas.forEach(function (validationMetadata) {
                if (validationMetadata.propertyName === key && validationMetadata.type !== 'conditionalValidation') {
                    switch (validationMetadata.type) {
                        /*
                        case ValidationTypes.IS_NOT_EMPTY: {
                            formGroupField.push(Validators.required);
                            break;
                        }
                        case ValidationTypes.IS_EMAIL: {
                            formGroupField.push(Validators.required);
                            break;
                        }*/
                        default: {
                            for (var typeKey in ValidationTypes) {
                                if (ValidationTypes.hasOwnProperty(typeKey)) {
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] === undefined &&
                                        validationMetadata.type === 'nestedValidation' &&
                                        typeKey === 'NESTED_VALIDATION') {
                                        /** @type {?} */
                                        var nestedValidate = function (c) {
                                            /** @type {?} */
                                            var isValid = (c.parent && c.parent.value) ?
                                                validateSync(c.value, validatorOptions).length === 0 : true;
                                            return isValid ? null : {
                                                nestedValidate: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(nestedValidate);
                                    }
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] === undefined &&
                                        validationMetadata.type === 'customValidation' &&
                                        typeKey === 'CUSTOM_VALIDATION') {
                                        /** @type {?} */
                                        var customValidation = function (c) {
                                            // todo: refactor
                                            /** @type {?} */
                                            var object = (c.parent instanceof DynamicFormGroup) ?
                                                ((/** @type {?} */ (c.parent))).object :
                                                (c.parent ? c.parent.value : {});
                                            if (object) {
                                                object[key] = c.value;
                                            }
                                            /** @type {?} */
                                            var validateErrors = (c.parent && c.parent.value) ?
                                                validateSync(object, validatorOptions) : [];
                                            /** @type {?} */
                                            var isValid = validateErrors.filter(function (error) {
                                                if (error.children.length &&
                                                    error.children.filter(function (children) { return children.property === key; })) {
                                                    return true;
                                                }
                                                return true;
                                            }).length === 0;
                                            return isValid ? null : {
                                                customValidation: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(customValidation);
                                    }
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] !== undefined) {
                                        /** @type {?} */
                                        var customValidate = function (c) {
                                            if (!c) {
                                                return null;
                                            }
                                            /** @type {?} */
                                            var isValid = c.parent && c.parent.value ? validator.validateValueByMetadata(c.value, validationMetadata) : true;
                                            if (!isValid && conditionalValidations.length > 0) {
                                                // todo: refactor
                                                /** @type {?} */
                                                var object = (c.parent instanceof DynamicFormGroup) ?
                                                    ((/** @type {?} */ (c.parent))).object :
                                                    (c.parent ? c.parent.value : {});
                                                if (object) {
                                                    object[key] = c.value;
                                                }
                                                /** @type {?} */
                                                var validateErrors = (c.parent && c.parent.value) ?
                                                    validateSync(object, validatorOptions) : [];
                                                isValid = validateErrors.filter(function (error) {
                                                    if (error.property === key) {
                                                        return true;
                                                    }
                                                    return false;
                                                }).length === 0;
                                            }
                                            return isValid ? null : {
                                                customValidate: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(customValidate);
                                    }
                                }
                            }
                            break;
                        }
                    }
                }
            });
            if (formGroupField[0] instanceof DynamicFormGroup) {
                formGroupFields[key] = formGroupField[0];
            }
            else {
                formGroupFields[key] = [
                    formGroupField[0],
                    formGroupField.filter(function (item, index) { return index !== 0; })
                ];
            }
        });
        return formGroupFields;
    };
    /**
     * @private
     * @param {?} fields
     * @return {?}
     */
    DynamicFormGroup.prototype.onlyFields = /**
     * @private
     * @param {?} fields
     * @return {?}
     */
    function (fields) {
        var _this = this;
        /** @type {?} */
        var newFields = {};
        if (fields !== undefined) {
            Object.keys(fields).forEach(function (key) {
                if (fields[key] instanceof DynamicFormGroup) {
                    newFields[key] = _this.onlyFields(((/** @type {?} */ (fields[key]))).fields);
                }
                else {
                    newFields[key] = fields[key];
                }
            });
        }
        return newFields;
    };
    Object.defineProperty(DynamicFormGroup.prototype, "object", {
        get: /**
         * @return {?}
         */
        function () {
            var _this = this;
            /** @type {?} */
            var object = this._object ? this.classToClass(this._object) : new this.factoryModel();
            if (object !== undefined) {
                Object.keys(this.controls).forEach(function (key) {
                    if (_this.controls[key] instanceof DynamicFormGroup) {
                        object[key] = ((/** @type {?} */ (_this.controls[key]))).object;
                    }
                    else {
                        object[key] = _this.controls[key].value;
                    }
                });
            }
            return this.plainToClass(this.factoryModel, object);
        },
        set: /**
         * @param {?} object
         * @return {?}
         */
        function (object) {
            var _this = this;
            if (object instanceof this.factoryModel) {
                this._object = this.classToClass(object);
            }
            else {
                this._object = this.plainToClass(this.factoryModel, (/** @type {?} */ (object)));
            }
            Object.keys(this.controls).forEach(function (key) {
                if (_this.controls[key] instanceof DynamicFormGroup) {
                    ((/** @type {?} */ (_this.controls[key]))).object = _this._object ? _this._object[key] : {};
                }
                else {
                    _this.controls[key].setValue((_this._object && _this._object[key]) ? _this._object[key] : undefined);
                }
            });
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @private
     * @param {?} errors
     * @return {?}
     */
    DynamicFormGroup.prototype.transformValidationErrors = /**
     * @private
     * @param {?} errors
     * @return {?}
     */
    function (errors) {
        var _this = this;
        /** @type {?} */
        var customErrors = {};
        errors.forEach(function (error) {
            if (error && error.constraints !== undefined) {
                Object.keys(error.constraints).forEach(function (key) {
                    if (!customErrors[error.property]) {
                        customErrors[error.property] = [];
                    }
                    if (((/** @type {?} */ (customErrors[error.property]))).indexOf(error.constraints[key]) === -1) {
                        ((/** @type {?} */ (customErrors[error.property]))).push(error.constraints[key]);
                    }
                });
            }
            if (error.children !== undefined && error.children.length) {
                customErrors[error.property] = _this.transformValidationErrors(error.children);
            }
        });
        return customErrors;
    };
    /**
     * @template TClassModel
     * @param {?} object
     * @return {?}
     */
    DynamicFormGroup.prototype.classToClass = /**
     * @template TClassModel
     * @param {?} object
     * @return {?}
     */
    function (object) {
        return classToClass(object, { ignoreDecorators: true });
    };
    /**
     * @template TClassModel, Object
     * @param {?} cls
     * @param {?} plain
     * @return {?}
     */
    DynamicFormGroup.prototype.plainToClass = /**
     * @template TClassModel, Object
     * @param {?} cls
     * @param {?} plain
     * @return {?}
     */
    function (cls, plain) {
        return plainToClass(cls, plain, { ignoreDecorators: true });
    };
    /**
     * @private
     * @param {?=} errors
     * @param {?=} externalErrors
     * @return {?}
     */
    DynamicFormGroup.prototype.mergeErrors = /**
     * @private
     * @param {?=} errors
     * @param {?=} externalErrors
     * @return {?}
     */
    function (errors, externalErrors) {
        return mergeWith(errors, externalErrors, function (objValue, srcValue) {
            if (Array.isArray(objValue) &&
                Array.isArray(srcValue) &&
                objValue.filter(function (objItem) { return srcValue.indexOf(objItem) !== -1; }).length === 0) {
                return objValue.concat(srcValue);
            }
        });
    };
    /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    DynamicFormGroup.prototype.validate = /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    function (externalErrors, validatorOptions) {
        if (externalErrors === undefined) {
            externalErrors = cloneDeep(this.externalErrors);
        }
        if (validatorOptions === undefined) {
            validatorOptions = cloneDeep(this.validatorOptions);
        }
        if (!externalErrors) {
            externalErrors = {};
        }
        /** @type {?} */
        var errors = validateSync(this.object, validatorOptions);
        /** @type {?} */
        var transformedErrors = this.transformValidationErrors(errors);
        /** @type {?} */
        var allErrors = this.mergeErrors(externalErrors, transformedErrors);
        this.markAsInvalidForExternalErrors(externalErrors, this.controls);
        this.customValidateErrors.next(allErrors);
    };
    /**
     * @private
     * @param {?} errors
     * @param {?} controls
     * @return {?}
     */
    DynamicFormGroup.prototype.markAsInvalidForExternalErrors = /**
     * @private
     * @param {?} errors
     * @param {?} controls
     * @return {?}
     */
    function (errors, controls) {
        Object.keys(controls).forEach(function (field) {
            /** @type {?} */
            var control = controls[field];
            if (control instanceof FormControl) {
                if (errors && errors[field]) {
                    control.setErrors({ 'externalError': true });
                }
                else {
                    if (control.errors && control.errors.externalError === true) {
                        control.setErrors(null);
                    }
                }
            }
            else if (control instanceof DynamicFormGroup) {
                control.markAsInvalidForExternalErrors(errors && errors[field] ? (/** @type {?} */ (errors[field])) : {}, control.controls);
            }
        });
    };
    /**
     * @return {?}
     */
    DynamicFormGroup.prototype.validateAllFormFields = /**
     * @return {?}
     */
    function () {
        var _this = this;
        Object.keys(this.controls).forEach(function (field) {
            /** @type {?} */
            var control = _this.get(field);
            if (control instanceof FormControl) {
                control.markAsTouched({ onlySelf: true });
            }
            else if (control instanceof DynamicFormGroup) {
                control.validateAllFormFields();
            }
        });
    };
    return DynamicFormGroup;
}(FormGroup));

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var DynamicFormBuilder = /** @class */ (function (_super) {
    __extends(DynamicFormBuilder, _super);
    function DynamicFormBuilder() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} controlsConfig
     * @param {?=} extra
     * @return {?}
     */
    DynamicFormBuilder.prototype.group = /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} controlsConfig
     * @param {?=} extra
     * @return {?}
     */
    function (factoryModel, controlsConfig, extra) {
        var _this = this;
        if (extra !== undefined && extra.customValidatorOptions === undefined) {
            extra.customValidatorOptions = { validationError: { target: false } };
        }
        /** @type {?} */
        var newControlsConfig;
        if (controlsConfig !== undefined) {
            newControlsConfig = controlsConfig;
        }
        // experimental
        if (controlsConfig === undefined) {
            newControlsConfig = new factoryModel();
            Object.keys(newControlsConfig).forEach(function (key) {
                if (newControlsConfig[key] &&
                    newControlsConfig[key].constructor &&
                    typeof newControlsConfig[key] === 'object' &&
                    (newControlsConfig[key].length === undefined ||
                        (newControlsConfig[key].length !== undefined &&
                            Object.keys(newControlsConfig[key].length).length === newControlsConfig[key].length))) {
                    newControlsConfig[key] = _this.group(newControlsConfig[key].constructor, undefined, extra);
                }
            });
        }
        /** @type {?} */
        var dynamicFormGroup = new DynamicFormGroup(factoryModel, newControlsConfig);
        /** @type {?} */
        var formGroup = _super.prototype.group.call(this, DynamicFormGroup.getClassValidators(factoryModel, newControlsConfig, (extra && extra.customValidatorOptions) ? extra.customValidatorOptions : undefined), extra);
        Object.keys(formGroup.controls).forEach(function (key) {
            dynamicFormGroup.addControl(key, formGroup.controls[key]);
        });
        dynamicFormGroup.valueChanges.subscribe(function (data) {
            dynamicFormGroup.validate(undefined, (extra && extra.customValidatorOptions) ? extra.customValidatorOptions : undefined);
        });
        return dynamicFormGroup;
    };
    return DynamicFormBuilder;
}(FormBuilder));

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { DynamicFormBuilder, DynamicFormGroup };

//# sourceMappingURL=ngx-dynamic-form-builder.js.map